@@ -1,9 +1,9 @@
-;;; gnorb-bbdb.el --- The BBDB-centric functions of gnorb
+;;; gnorb-bbdb.el --- The BBDB-centric functions of gnorb  -*- lexical-binding: t -*-
 
 ;; Copyright (C) 2014  Free Software Foundation, Inc.
 
 ;; Author: Eric Abrahamsen <eric@ericabrahamsen.net>
-;; Keywords: 
+;; Keywords:
 
 ;; This program is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -20,14 +20,16 @@
 
 ;;; Commentary:
 
-;; 
+;; The Gnorb package has no hard dependency on BBDB, so you'll have to
+;; install it manually.  Gnorb is compatible with whichever version of
+;; BBDB is current in the Emacs package manager.  I believe it comes
+;; from Melpa.
 
 ;;; Code:
 
-(eval-when-compile
-  (require 'cl))
-
 (require 'bbdb)
+(require 'bbdb-com)
+(require 'bbdb-mua)
 (require 'gnorb-utils)
 (require 'cl-lib)
 
@@ -69,7 +71,7 @@ Setting it to the symbol seen will collect the messages most
 recently opened and viewed. The symbol received means gnorb will
 collect the most recent messages by Date header.
 
-In other words, if this variable is set to 'received, and a
+In other words, if this variable is set to `received', and a
 record's messages field is already full of recently-received
 messages, opening a five-year-old message (for instance) from
 this record will not push a link to the message into the field."
@@ -92,9 +94,9 @@ mentioned in the docstring of `format-time-string', which see."
 (defcustom gnorb-bbdb-message-link-format-one "%:count"
   "How a single message is formatted in the list of recent messages.
 This format string is used in single-line display -- note that by
-default, no user-created xfields are displayed in the 'one-line
+default, no user-created xfields are displayed in the `one-line'
 layout found in `bbdb-layout-alist'. If you want this field to
-appear there, put its name in the \"order\" list of the 'one-line
+appear there, put its name in the \"order\" list of the `one-line'
 layout.
 
 Available information for each message includes the subject, the
@@ -105,7 +107,7 @@ mentioned in the docstring of `format-time-string', which see."
   :group 'gnorb-bbdb
   :type 'string)
 
-(defface gnorb-bbdb-link (org-compatible-face 'org-link nil)
+(defface gnorb-bbdb-link '((t :inherit org-link))
   "Custom face for displaying message links in the *BBDB* buffer.
   Defaults to org-link."
   :group 'gnorb-bbdb)
@@ -148,12 +150,10 @@ returns either t or nil. In this case, the second element of the
 list is disregarded.
 
 All following elements should be field setters for the message to
-be composed, just as in `gnus-posting-styles'.
+be composed, just as in `gnus-posting-styles'."
 
-An example value might look like:"
-  :group 'gnorb-bbdb)
-
-(defvar message-mode-hook)
+  :group 'gnorb-bbdb
+  :type 'list)
 
 (when (fboundp 'bbdb-record-xfield-string)
   (fset (intern (format "bbdb-read-xfield-%s"
@@ -163,8 +163,8 @@ An example value might look like:"
 
   (fset (intern (format "bbdb-display-%s-multi-line"
 			gnorb-bbdb-org-tag-field))
-	(lambda (record)
-	  (gnorb-bbdb-display-org-tags record))))
+	(lambda (record indent)
+	  (gnorb-bbdb-display-org-tags record indent))))
 
 (defun gnorb-bbdb-read-org-tags (&optional init)
   "Read Org mode tags, with `completing-read-multiple'."
@@ -176,12 +176,12 @@ An example value might look like:"
 		     "[ \t\n]*"))
 	    (crm-local-completion-map bbdb-crm-local-completion-map)
 	    (table (cl-mapcar #'car
-			   (org-global-tags-completion-table
-			    (org-agenda-files))))
+			      (org-global-tags-completion-table
+			       (org-agenda-files))))
 	    (init (if (consp init)
-		      (bbdb-join init
-				 (nth 2 (assq gnorb-bbdb-org-tag-field
-					      bbdb-separator-alist)))
+		      (apply #'bbdb-concat (nth 2 (assq gnorb-bbdb-org-tag-field
+							bbdb-separator-alist))
+			     init)
 		    init)))
 	(completing-read-multiple
 	 "Tags: " table
@@ -189,7 +189,7 @@ An example value might look like:"
     (bbdb-split gnorb-bbdb-org-tag-field
 		(bbdb-read-string "Tags: " init))))
 
-(defun gnorb-bbdb-display-org-tags (record)
+(defun gnorb-bbdb-display-org-tags (record indent)
   "Display the Org tags associated with the record.
 
 Org tags are stored in the `gnorb-bbdb-org-tags-field'."
@@ -199,16 +199,16 @@ Org tags are stored in the `gnorb-bbdb-org-tags-field'."
 	      record
 	      gnorb-bbdb-org-tag-field)))
     (when val
-      ;; We already know that `fmt' and `indent' are dynamically
-      ;; bound, shut up about it.
-      (with-no-warnings
-       (bbdb-display-text (format fmt gnorb-bbdb-org-tag-field)
-			  `(xfields ,full-field field-label)
-			  'bbdb-field-name)
-       (if (consp val)
-	   (bbdb-display-list val gnorb-bbdb-org-tag-field "\n")
-	 (insert
-	  (bbdb-indent-string (concat val "\n") indent)))))))
+      (bbdb-display-text (format (format " %%%ds: " (- indent 3))
+				 gnorb-bbdb-org-tag-field)
+			 `(xfields ,full-field field-label)
+			 'bbdb-field-name)
+      (if (consp val)
+	  (bbdb-display-list val gnorb-bbdb-org-tag-field "\n")
+	(insert
+	 (bbdb-indent-string (concat val "\n") indent))))))
+
+(defvar message-mode-hook)
 
 ;;;###autoload
 (defun gnorb-bbdb-mail (records &optional subject n verbose)
@@ -247,7 +247,7 @@ is non-nil (as in interactive calls) be verbose."
 (defun gnorb-bbdb-configure-posting-styles (recs)
   ;; My most magnificent work of copy pasta!
   (dolist (r recs)
-    (let (field val label rec-val element filep
+    (let (field val label rec-val filep
 		element v value results name address)
       (dolist (style gnorb-bbdb-posting-styles)
 	(setq field (pop style)
@@ -258,24 +258,34 @@ is non-nil (as in interactive calls) be verbose."
 	(unless (fboundp field)
 	  ;; what's the record's existing value for this field?
 	  (setq rec-val (bbdb-record-field r field)))
-	(when (cond
-	       ((eq field 'address)
-		(dolist (a rec-val)
-		  (unless (and label
-			       (not (string-match label (car a))))
-		    (string-match val (bbdb-format-address-default a)))))
-	       ((eq field 'phone)
-		(dolist (p rec-val)
-		  (unless (and label
-			       (not (string-match label (car p))))
-		    (string-match val (bbdb-phone-string p)))))
-	       ((consp rec-val)
-		(dolist (f rec-val)
-		  (string-match val f)))
-	       ((fboundp field)
-		(funcall field r))
-	       ((stringp rec-val)
-		(string-match val rec-val)))
+	(when (catch 'match
+		(cond
+		 ((eq field 'address)
+		  (dolist (a rec-val)
+		    (unless (and label
+				 (not (string-match label (car a))))
+		      (when
+			  (string-match-p
+			   val
+			   (bbdb-format-address-default a))
+			(throw 'match t)))))
+		 ((eq field 'phone)
+		  (dolist (p rec-val)
+		    (unless (and label
+				 (not (string-match label (car p))))
+		      (when
+			  (string-match-p val (bbdb-phone-string p))
+			(throw 'match t)))))
+		 ((consp rec-val)
+		  (dolist (f rec-val)
+		    (when (string-match-p val f)
+		      (throw 'match t))))
+		 ((fboundp field)
+		  (when (string-match-p (funcall field r))
+		    (throw 'match t)))
+		 ((stringp rec-val)
+		  (when (string-match-p val rec-val)
+		    (throw 'match t)))))
 	  ;; there are matches, run through the field setters in last
 	  ;; element of the sexp
 	  (dolist (attribute style)
@@ -326,7 +336,6 @@ is non-nil (as in interactive calls) be verbose."
       (setq name (assq 'name results)
 	    address (assq 'address results))
       (setq results (delq name (delq address results)))
-      (gnus-make-local-hook 'message-setup-hook)
       (setq results (sort results (lambda (x y)
 				    (string-lessp (car x) (car y)))))
       (dolist (result results)
@@ -395,15 +404,17 @@ both, use \"C-u\" before the \"*\"."
 	 (mapconcat
 	  'identity
 	  (delete-dups
-	   (cl-mapcan (lambda (r)
-		     (bbdb-record-xfield-split r gnorb-bbdb-org-tag-field))
-		   records)))
+	   (cl-mapcan
+	    (lambda (r)
+	      (bbdb-record-xfield-split r gnorb-bbdb-org-tag-field))
+	    records))
 	  "|")))
     (if tag-string
 	;; C-u = all headings, not just todos
-	(org-tags-view (not (equal current-prefix-arg '(4)))
-                       tag-string)
-      (error "No org-tags field present")))
+	(if (equal current-prefix-arg '(4))
+	    (org-tags-view nil tag-string)
+	  (org-tags-view t tag-string))
+      (error "No org-tags field present"))))
 
 ;;;###autoload
 (defun gnorb-bbdb-mail-search (records)
@@ -427,24 +438,23 @@ a prefix arg and \"*\", the prefix arg must come first."
     (when (equal current-prefix-arg '(4))
       (setq search-string
 	    (read-from-minibuffer
-	     (format "%s search string: " (cl-first backend)) search-string)))
+	     (format "%s search string: " (first backend)) search-string)))
     (funcall (cl-third backend) search-string)
-    (delete-other-windows)))  
+    (delete-other-windows)))
 
 ;;;###autoload
 (defun gnorb-bbdb-cite-contact (rec)
-  (interactive (list (gnorb-prompt-for-bbdb-record)))
+  (interactive (list (bbdb-completing-read-record "Record: ")))
   (let ((mail-string (bbdb-dwim-mail rec)))
    (if (called-interactively-p 'any)
        (insert mail-string)
      mail-string)))
 
 ;;; Field containing links to recent messages
-
 (when (boundp 'bbdb-xfield-label-list)
-  (add-to-list 'bbdb-xfield-label-list gnorb-bbdb-messages-field nil 'eq))
+ (add-to-list 'bbdb-xfield-label-list gnorb-bbdb-messages-field nil 'eq))
 
-(defun gnorb-bbdb-display-messages (record format)
+(defun gnorb-bbdb-display-messages (record format &optional indent)
   "Show links to the messages collected in the
 `gnorb-bbdb-messages-field' field of a BBDB record. Each link
 will be formatted using the format string in
@@ -460,14 +470,13 @@ layout type."
     (define-key map (kbd "<RET>") 'gnorb-bbdb-RET-open-link)
     (when val
       (when (eq format 'multi)
-	(with-no-warnings ; For `fmt'
-	  (bbdb-display-text (format fmt gnorb-bbdb-messages-field)
-			     `(xfields ,full-field field-label)
-			     'bbdb-field-name)))
+	(bbdb-display-text (format (format " %%%ds: " (- indent 3))
+				   gnorb-bbdb-messages-field)
+			   `(xfields ,full-field field-label)
+			   'bbdb-field-name))
       (insert (cond ((and (stringp val)
 			  (eq format 'multi))
-		     (with-no-warnings ; For `indent'
-		       (bbdb-indent-string (concat val "\n") indent)))
+		     (bbdb-indent-string (concat val "\n") indent))
 		    ((listp val)
 		     ;; Why aren't I using `bbdb-display-list' with a
 		     ;; preformatted list of messages?
@@ -502,8 +511,8 @@ layout type."
 
 (fset (intern (format "bbdb-display-%s-multi-line"
 		      gnorb-bbdb-messages-field))
-      (lambda (record)
-	(gnorb-bbdb-display-messages record 'multi)))
+      (lambda (record indent)
+	(gnorb-bbdb-display-messages record 'multi indent)))
 
 (fset (intern (format "bbdb-display-%s-one-line"
 		      gnorb-bbdb-messages-field))
@@ -514,7 +523,7 @@ layout type."
 
 (fset (intern (format "bbdb-read-xfield-%s"
 		      gnorb-bbdb-messages-field))
-      (lambda (&optional init)
+      (lambda (&optional _init)
 	(user-error "This field shouldn't be edited manually")))
 
 ;; Open links from the *BBDB* buffer.
@@ -537,8 +546,7 @@ that contact will start collecting links to messages."
 		current-prefix-arg))
   (unless (fboundp 'bbdb-record-xfield-string)
     (user-error "This function only works with the git version of BBDB"))
-  (let* ((record (bbdb-current-record))
-	 msg-list target-msg)
+  (let (msg-list target-msg)
     (if (not (memq gnorb-bbdb-messages-field
 		   (mapcar 'car (bbdb-record-xfields record))))
 	(when (y-or-n-p
@@ -597,16 +605,10 @@ to a message into the record's `gnorb-bbdb-messages-field'."
 			  (parse-time-string (mail-header-date heads))))
 	     (subject (mail-header-subject heads))
 	     (id (mail-header-id heads))
-	     (group gnus-newsgroup-name)
+	     (group (gnorb-get-real-group-name
+		     gnus-newsgroup-name
+		     art-no))
 	     link)
-	;; check for both nnvirtual and nnir, and link to the real
-	;; group in those cases
-	(when (eq (car (gnus-find-method-for-group group))
-		  'nnvirtual)
-	  (setq group (car (nnvirtual-map-article art-no))))
-	(when (eq (car (gnus-find-method-for-group group))
-		  'nnir)
-	  (setq group (nnir-article-group art-no)))
 	(if (not (and date subject id group))
 	    (message "Could not save a link to this message")
 	  (setq link (make-gnorb-bbdb-link :subject subject :date date
@@ -620,7 +622,7 @@ to a message into the record's `gnorb-bbdb-messages-field'."
 			      (time-less-p
 			       (gnorb-bbdb-link-date b)
 			       (gnorb-bbdb-link-date a))))))
-	  (setq val (cl-subseq val 0 gnorb-bbdb-collect-N-messages))
+	  (setq val (cl-subseq val 0 (min (length val) gnorb-bbdb-collect-N-messages)))
 	  (bbdb-record-set-xfield record
 				  gnorb-bbdb-messages-field
 				  (delq nil val))
