@@ -1,4 +1,4 @@
-;;; gnorb-org.el --- The Org-centric functions of gnorb
+;;; gnorb-org.el --- The Org-centric functions of gnorb -*- lexical-binding: t -*-
 
 ;; Copyright (C) 2014  Free Software Foundation, Inc.
 
@@ -24,11 +24,26 @@
 
 ;;; Code:
 
-(eval-when-compile
-  (require 'cl))
-
 (require 'gnorb-utils)
-(require 'cl-lib)
+(eval-when-compile (require 'cl-lib))
+
+(defvar gnorb-bbdb-posting-styles)
+(defvar gnorb-bbdb-org-tag-field)
+(defvar bbdb-buffer-name)
+(defvar message-alternative-emails)
+
+;; This many autoloads means either we should require bbdb outright,
+;; or something needs refactoring.
+(autoload 'gnorb-bbdb-configure-posting-styles "gnorb-bbdb")
+(autoload 'gnorb-registry-org-id-search "gnorb-registry")
+(autoload 'bbdb-completing-read-record "bbdb")
+(autoload 'bbdb-record-name "bbdb")
+(autoload 'bbdb-message-search "bbdb")
+(autoload 'bbdb-mail-address "bbdb")
+(autoload 'bbdb-record-xfield "bbdb")
+(autoload 'bbdb-records "bbdb")
+(autoload 'bbdb-search "bbdb")
+(autoload 'bbdb-display-records "bbdb")
 
 (defgroup gnorb-org nil
   "The Org bits of Gnorb."
@@ -42,13 +57,12 @@
   :type 'hook)
 
 (defcustom gnorb-org-trigger-actions
-  '(("todo state" . todo)
-    ("take note" . note)
-    ("don't associate" . no-associate)
-    ("only associate" . associate)
-;    ("capture to child" . cap-child)
-;    ("capture to sibling" . cap-sib)
-)
+  '((?t "todo state" todo)
+    (?n "take note" note)
+    (?d "don't associate" no-associate)
+    (?o "only associate" associate)
+    (?c "capture to child" cap-child)
+    (?s "capture to sibling" cap-sib))
   "List of potential actions that can be taken on headings.
 
 When triggering an Org heading after receiving or sending a
@@ -59,8 +73,8 @@ todo state: Associate the message, and change TODO state.
 take note: Associate the message, and take a note.
 don't associate: Do nothing at all, don't connect the message and TODO.
 only associate: Associate the message with this heading, do nothing else.
-capture to child: [not yet implemented] Associate this message with a new child heading.
-capture to sibling: [not yet implemented] Associate this message with a new sibling heading.
+capture to child: Associate this message with a new child heading.
+capture to sibling: Associate this message with a new sibling heading.
 
 You can reorder this list or remove items as suits your workflow.
 The two \"capture\" options will use the value of
@@ -68,12 +82,14 @@ The two \"capture\" options will use the value of
 template.
 
 You can also add custom actions to the list. Actions should be a
-cons of a string tag and a symbol indicating a custom function.
-This function will be called on the heading in question, and
-passed a plist containing information about the message from
-which we're triggering."
+list of three elements: a character key, a string tag and a
+symbol indicating a custom function.  The custom function will be
+called on the heading in question, and passed a plist containing
+information about the message from which we're triggering."
+
   :group 'gnorb-org
-  :type 'list)
+  :type 'list
+  :package-version '(gnorb . "1.1.3"))
 
 (defcustom gnorb-org-msg-id-key "GNORB_MSG_ID"
   "The name of the org property used to store the Message-IDs
@@ -85,6 +101,8 @@ which we're triggering."
 (defcustom gnorb-org-mail-scan-scope 2
   "Number of paragraphs to scan for mail-related links.
 
+Or set to 'all to scan the whole subtree.
+
 When handling a TODO heading with `gnorb-org-handle-mail', Gnorb
 will typically reply to the most recent message associated with
 this heading. If there are no such messages, or message tracking
@@ -92,7 +110,10 @@ is disabled entirely, or `gnorb-org-handle-mail' has been called
 with a prefix arg, the heading and body text of the subtree under
 point will instead be scanned for gnus:, mailto:, and bbdb:
 links. This option controls how many paragraphs of body text to
-scan. Set to 0 to only look in the heading.")
+scan. Set to 0 to only look in the heading."
+  :group 'gnorb-org
+  :type '(choice (const :tag "Whole subtree" all)
+		 (integer :tag "Number of paragraphs")))
 
 (make-obsolete-variable
  'gnorb-org-mail-scan-strategies
@@ -127,7 +148,7 @@ There's really no reason to use this instead of regular old
 `org-insert-link' with BBDB completion. But there might be in the
 future!"
   ;; this needs to handle an active region.
-  (interactive (list (gnorb-prompt-for-bbdb-record)))
+  (interactive (list (bbdb-completing-read-record "Record: ")))
   (let* ((name (bbdb-record-name rec))
 	 (link (concat "bbdb:" (org-link-escape name))))
     (org-store-link-props :type "bbdb" :name name
@@ -150,7 +171,7 @@ we came from."
   (setq gnorb-message-org-ids nil)
   (gnorb-restore-layout))
 
-(defun gnorb-org-extract-links (&optional arg region)
+(defun gnorb-org-extract-links (&optional _arg region)
   "See if there are viable links in the subtree under point."
   ;; We're not currently using the arg. What could we do with it?
   (let (strings)
@@ -168,7 +189,7 @@ we came from."
 		 :raw-value
 		 head)
 		strings)
-	  (org-element-map tree 'paragraph
+	  (org-element-map tree '(paragraph drawer)
 	    (lambda (p)
 	      (push (org-element-interpret-data p)
 		    strings))
@@ -180,9 +201,11 @@ we came from."
 	    (cond ((eq gnorb-org-mail-scan-scope 'all)
 		   strings)
 		  ((numberp gnorb-org-mail-scan-scope)
-		   (delq nil
-			 (cl-subseq
-			  strings 0 (1+ gnorb-org-mail-scan-scope))))
+		   (cl-subseq
+		    (reverse strings)
+		    0 (min
+		       (length strings)
+		       (1+ gnorb-org-mail-scan-scope))))
 		  ;; We could provide more options here. 'tree vs
 		  ;; 'subtree, for instance.
 		  (t
@@ -210,6 +233,17 @@ See the docstring of `gnorb-org-handle-mail' for details."
 	     (gnorb-collect-ids)))))
       (gnorb-org-extract-mail-tracking assoc-msg-ids arg region))))
 
+(defun gnorb-user-address-match-p (addr)
+  "Return t if ADDR seems to match the user's email address."
+  (cond
+   ((stringp message-alternative-emails)
+    (string-match-p message-alternative-emails
+		    addr))
+   ((functionp message-alternative-emails)
+    (funcall message-alternative-emails addr))
+   (user-mail-address
+    (string-match-p user-mail-address addr))))
+
 (defun gnorb-org-extract-mail-tracking (assoc-msg-ids &optional arg region)
 
   (let* ((all-links (gnorb-org-extract-links nil region))
@@ -222,11 +256,8 @@ See the docstring of `gnorb-org-handle-mail' for details."
 	      (cl-remove-if
 	       (lambda (m)
 		 (let ((from (car (gnus-registry-get-id-key m 'sender))))
-		   (or (null from)
-		       (string-match-p
-			user-mail-address from)
-		       (string-match-p
-			message-alternative-emails from))))
+		   (and from
+			(null (gnorb-user-address-match-p from)))))
 	       assoc-msg-ids)
 	      (lambda (r l)
 		(time-less-p
@@ -274,10 +305,14 @@ headings."
     (when messages
       (insert ", "))
     (insert (mapconcat 'identity mails ", ")))
-  ;; Return us after message is sent.
-  (add-to-list 'message-exit-actions
-	       'gnorb-org-restore-after-send t)
-  ;; Set headers from MAIL_* properties (from, cc, and bcc).
+  ;; Commenting this out because
+  ;; `gnorb-gnus-check-outgoing-headers' is set unconditionally in the
+  ;; `message-send-hook, so this should be redundant.  Also, we've
+  ;; switched to using message-send-actions.
+  
+  ;; (add-to-list
+  ;; 'message-exit-actions 'gnorb-org-restore-after-send t) Set
+  ;; headers from MAIL_* properties (from, cc, and bcc).
   (cl-flet ((sh (h)
 		(when (cdr h)
 		  (funcall (intern (format "message-goto-%s" (car h))))
@@ -301,7 +336,7 @@ headings."
   ;; insert text, if any
   (when text
     (message-goto-body)
-    (insert"\n")
+    (insert "\n")
     (if (bufferp text)
 	(insert-buffer-substring text)
       (insert text)))
@@ -409,7 +444,7 @@ around that link before you call this message."
 	       (message-mode-hook (copy-sequence message-mode-hook))
 	       mails)
 	  (when file
-	    (cons file attachments))
+	    (setq attachments (cons file attachments)))
 	  (when recs
 	    (setq recs
 		  (delq nil
@@ -478,7 +513,9 @@ async, subtreep, visible-only, and body-only."
   "Correspondence between export backends and their
 respective (usual) file extensions. Ugly way to do it, but what
 the hey..."
-  :group 'gnorb-org)
+  :group 'gnorb-org
+  :type '(repeat
+	  (list symbol string)))
 
 (defvar org-export-show-temporary-export-buffer)
 
@@ -501,13 +538,17 @@ default set of parameters."
   ;; got too much hard-coded stuff.
   (interactive "P")
   (org-back-to-heading t)
-  (let* ((backend-string
+  (let* ((bkend-var
+	  (if (boundp 'org-export--registered-backends)
+	      org-export--registered-backends
+	    org-export-registered-backends))
+	 (backend-string
 	  (org-completing-read
 	   "Export backend: "
 	   (mapcar (lambda (b)
 		     (symbol-name (org-export-backend-name b)))
-		   org-export--registered-backends)
-           nil t))
+		   bkend-var)
+	   nil t))
 	 (backend-symbol (intern backend-string))
 	 (f-or-t (org-completing-read "Export as file or text? "
 				      '("file" "text") nil t))
@@ -522,16 +563,23 @@ default set of parameters."
 		       "*Gnorb Export*"
 		       ,@opts
 		       ,gnorb-org-email-subtree-text-parameters))
-	    (apply 'org-export-to-file
-		   `(,backend-symbol
-		     ,(org-export-output-file-name
-		       (second (assoc backend-symbol gnorb-org-export-extensions))
-		       t gnorb-tmp-dir)
-		     ,@opts
-		     ,gnorb-org-email-subtree-file-parameters))))
+	    (if (eq backend-symbol 'odt)
+		;; Need to special-case odt output, as it does too
+		;; many clever things.  The only downside to this is
+		;; it's impossible to put the exported file in the
+		;; /tmp/ directory -- it will go wherever it would
+		;; have gone with manual export.
+		(apply #'org-odt-export-to-odt
+		       (append (cl-subseq gnorb-org-email-subtree-file-options 0 3)
+			       (list gnorb-org-email-subtree-file-parameters)))
+	     (apply 'org-export-to-file
+		    `(,backend-symbol
+		      ,(org-export-output-file-name
+			(cl-second (assoc backend-symbol gnorb-org-export-extensions))
+			t gnorb-tmp-dir)
+		      ,@opts
+		      ,gnorb-org-email-subtree-file-parameters)))))
 	 text file)
-    (setq gnorb-window-conf (current-window-configuration))
-    (move-marker gnorb-return-marker (point))
     (if (bufferp result)
 	(setq text result)
       (setq file result))
@@ -541,7 +589,8 @@ default set of parameters."
   "Should the capture process store a link to the gnus message or
   BBDB record under point, even if it's not part of the template?
   You'll probably end up needing it, anyway."
-  :group 'gnorb-org)
+  :group 'gnorb-org
+  :type 'boolean)
 
 (defun gnorb-org-capture-collect-link ()
   (when gnorb-org-capture-collect-link-p
@@ -564,7 +613,8 @@ default set of parameters."
 Records are considered matching if they have an `org-tags' field
 matching the current Agenda search. The name of that field can be
 customized with `gnorb-bbdb-org-tag-field'."
-  :group 'gnorb-org)
+  :group 'gnorb-org
+  :type 'boolean)
 
 (defcustom gnorb-org-bbdb-popup-layout 'pop-up-multi-line
   "Default BBDB buffer layout for automatic Org Agenda display."
@@ -593,7 +643,7 @@ search."
 		 (eq org-agenda-type 'tags))
 	    (or (called-interactively-p 'any)
 		gnorb-org-agenda-popup-bbdb))
-	   (let ((todo-only nil)
+	   (let ((org--matcher-tags-todo-only nil)
 		 (str (or str org-agenda-query-string))
 		 (re "^&?\\([-+:]\\)?\\({[^}]+}\\|LEVEL\\([<=>]\\{1,2\\}\\)\\([0-9]+\\)\\|\\(\\(?:[[:alnum:]_]+\\(?:\\\\-\\)*\\)+\\)\\([<>=]\\{1,2\\}\\)\\({[^}]+}\\|\"[^\"]*\"\\|-?[.0-9]+\\(?:[eE][-+]?[0-9]+\\)?\\)\\|[[:alnum:]_@#%]+\\)")
 		 or-terms term rest out-or acc tag-clause)
@@ -617,9 +667,12 @@ search."
 			(let ((rec-tags (bbdb-record-xfield
 					 r gnorb-bbdb-org-tag-field)))
 			  (and rec-tags
-			       (let ((tags-list (org-split-string rec-tags ":"))
+			       (let ((tags-list (if (stringp rec-tags)
+						    (org-split-string rec-tags ":")
+						  rec-tags))
 				     (case-fold-search t)
 				     (org-trust-scanner-tags t))
+				 ;; This is bad, we're lexically bound, now.
 				 (eval tag-clause)))))
 		      (bbdb-records))))))
 	  ((eq major-mode 'org-mode)
@@ -649,14 +702,17 @@ search."
 ;;; Groups from the gnorb gnus server backend
 
 ;;;###autoload
-(defun gnorb-org-view ()
+(defun gnorb-org-view (arg)
   "Search the subtree at point for links to gnus messages, and
-then show them in an ephemeral group, in gnus.
+then show them in an ephemeral group, in Gnus.
+
+With a prefix arg, create a search group that will persist across
+Gnus sessions, and can be refreshed.
 
 This won't work unless you've added a \"nngnorb\" server to
 your gnus select methods."
   ;; this should also work on the active region, if there is one.
-  (interactive)
+  (interactive "P")
   (require 'gnorb-gnus)
   (setq gnorb-window-conf (current-window-configuration))
   (move-marker gnorb-return-marker (point))
@@ -675,7 +731,10 @@ your gnus select methods."
       (org-back-to-heading)
       (setq id (concat "id+" (org-id-get-create)))
       (gnorb-gnus-search-messages
-       id
+       id arg
+       (replace-regexp-in-string
+	org-bracket-link-regexp "\\3"
+	(nth 4 (org-heading-components)))
        `(when (and (window-configuration-p gnorb-window-conf)
 		   gnorb-return-marker)
 	  (set-window-configuration gnorb-window-conf)
