@@ -1,4 +1,4 @@
-;;; gnorb-utils.el --- Common utilities for all gnorb stuff.
+;;; gnorb-utils.el --- Common utilities for all gnorb stuff -*- lexical-binding: t -*-
 
 ;; Copyright (C) 2014  Free Software Foundation, Inc.
 
@@ -24,14 +24,19 @@
 
 ;;; Code:
 
-(eval-when-compile
-  (require 'cl))
+(eval-when-compile (require 'cl-lib))
+(require 'pcase)
+(require 'org)
+(require 'org-agenda)
+(require 'org-element)
 
+(require 'mailcap)
 (mailcap-parse-mimetypes)
 
 (defgroup gnorb nil
   "Glue code between Gnus, Org, and BBDB."
-  :tag "Gnorb")
+  :tag "Gnorb"
+  :group 'mail)
 
 (make-obsolete-variable
  'gnorb-trigger-todo-default
@@ -39,21 +44,6 @@
 `gnorb-org-trigger-actions'"
  "September 8, 2014" 'set)
 
-(defun gnorb-prompt-for-bbdb-record ()
-  "Prompt the user for a BBDB record."
-  (let ((recs (bbdb-records))
-	name)
-    (while (> (length recs) 1)
-      (setq name
-	    (completing-read
-	     (format "Filter records by regexp (%d remaining): "
-		     (length recs))
-	     (mapcar 'bbdb-record-name recs)))
-      (setq recs (bbdb-search recs name name name nil nil)))
-    (if recs
-	(car recs)
-      (error "No matching records"))))
-
 (defvar gnorb-tmp-dir (make-temp-file "emacs-gnorb" t)
   "Temporary directory where attachments etc are saved.")
 
@@ -74,6 +64,11 @@ are sent, or Org headings triggered.")
   "Return point here after various actions, to be used together
 with `gnorb-window-conf'.")
 
+(defvar gnorb-trigger-capture-location nil
+  "Marker pointing at the location where we want to place capture
+  templates, for the capture-to-child and capture-to-sibling
+  trigger actions.")
+
 (defcustom gnorb-mail-header "X-Org-ID"
   "Name of the mail header used to store the ID of a related Org
   heading. Only used locally: always stripped when the mail is
@@ -170,6 +165,8 @@ link.
 
 3. Otherwise just follow the link as usual, in the current
 window."
+  (unless (gnus-alive-p)
+    (gnus))
   (let* ((sum-buffer (gnus-summary-buffer-name group))
 	 (target-buffer
 	  (cond
@@ -199,52 +196,146 @@ window."
   ;; We've probably already bracketed the id, but just in case this is
   ;; called from elsewhere...
   (let* ((id (gnorb-bracket-message-id id))
-	 (art-no (cdr (gnus-request-head id group)))
 	 (arts (gnus-group-unread group))
-	 success)
+	 artno success)
+    (or (setq artno (car (gnus-registry-get-id-key id 'artno)))
+	(progn
+	  (setq artno (cdr (gnus-request-head id group)))
+	  (gnus-registry-set-id-key id 'artno (list artno))))
     (gnus-activate-group group)
     (setq success (gnus-group-read-group arts t group))
     (if success
-	(gnus-summary-goto-article (or art-no id) nil t)
+	(gnus-summary-goto-article artno nil t)
       (signal 'error "Group could not be opened."))))
 
-(defun gnorb-trigger-todo-action (arg &optional id)
+;; I'd like to suggest this as a general addition to Emacs.  *Very*
+;; tired of abusing `completing-read' for this purpose.
+(defconst gnorb-select-valid-chars
+  (append (number-sequence 97 122)
+	  (number-sequence 65 90))
+  "A list of characters that are suitable for using as selection
+  keys.")
+
+(defvar gnorb-select-choice-buffer "*Selections*"
+  "The name of the buffer used to pop up selections.")
+
+(defun gnorb-select-from-list (prompt collection &optional key-func)
+  "Prompt the user to select something from COLLECTION.
+
+Selection can happen in a few different ways, depending on the
+nature of COLLECTION.  Its elements can be:
+
+1. A plain string.  Simply default to `completing-read'.
+
+2. (string object).  The function uses `completing-read' on the
+   strings, returning the selected object.
+
+3. (number object).  As above, but the user enters a number.
+
+4. (character string object).  As #3, but \"string\" is displayed
+   as a string label for object.
+
+5. (number string object).  As above, with numbers.
+
+COLLECTION can be passed in ready-made.  Alternately, KEY-FUNC
+can be provided.  The collection will be constructed by mapping
+this function over the list of objects, and then appending each
+object to the corresponding result.  In other words, KEY-FUNC
+should return one of the types above, minus the final \"object\"
+element.
+
+Alternately, KEY-FUNC can be the symbol 'char, in which case the
+elements of COLLECTION will automatically be keyed to ascending
+characters (52 or fewer), or 'number, which does the same with
+numbers (no upper bound)."
+  (interactive)
+  (let ((len (length collection)))
+    (cl-labels ((pop-up-selections
+		 (collection &optional charp)
+		 (pop-to-buffer gnorb-select-choice-buffer
+				'(display-buffer-in-side-window ((side . bottom))) t)
+		 (dolist (c collection)
+		   (insert (format "%s: %s\n"
+				   (if charp
+				       (char-to-string (car c))
+				     (car c))
+				   (nth 1 c))))))
+      (setq collection
+	    (pcase key-func
+	      ((pred null)
+	       collection)
+	      ('char
+	       (if (> len 52)
+		   (error "Use the char option with fewer than 52 items")
+		 ;; These distinctions between char/string
+		 ;; and number/char are totally manufactured.
+		 (seq-mapn #'list gnorb-select-valid-chars collection)))
+	      ('number
+	       (seq-mapn #'list (number-sequence 1 len) collection))
+	      ((and func (pred functionp))
+	       (seq-map (lambda (el)
+			  (let ((res (funcall func el)))
+			    (if (atom res)
+				(list res el)
+			      (append res
+				      (list el)))))
+			collection))
+	      (_ (error "Invalid key-func: %s" key-func))))
+      ;; We only test the car of collection to see what type it is.  If
+      ;; elements are mismatched, it's not our problem.
+      (unwind-protect
+	  (pcase (car collection)
+	    ((pred stringp)
+	     (completing-read prompt collection nil t))
+	    ((pred symbolp)
+	     (intern-soft (completing-read prompt collection nil t)))
+	    (`(,(pred stringp) ,_)
+	     (nth 1 (assoc (completing-read prompt collection nil t)
+			   collection)))
+	    ;; Looks like pcase might be the wrong tool for this job.
+	    ((or `(,(and c (pred numberp) (guard (memq c gnorb-select-valid-chars))) ,_)
+		 `(,(and c (pred numberp) (guard (memq c gnorb-select-valid-chars))) ,_ ,_))
+	     (pop-up-selections collection t)
+	     (car (last (assq (read-char
+			       (propertize prompt 'face 'minibuffer-prompt))
+			      collection))))
+	    ((or `(,(pred numberp) ,_)
+		 `(,(pred numberp) ,_ ,_))
+	     (pop-up-selections collection)
+	     (car (last (assq (read-number prompt)
+			      collection)))))
+	(when-let ((win (get-buffer-window gnorb-select-choice-buffer)))
+	  (quit-window win))))))
+
+(defun gnorb-trigger-todo-action (_arg &optional id)
   "Do the actual restore action. Two main things here. First: if
 we were in the agenda when this was called, then keep us in the
 agenda. Then let the user choose an action from the value of
 `gnorb-org-trigger-actions'."
-  (let ((agenda-p (eq major-mode 'org-agenda-mode))
-	(action (cdr (assoc
-		      (org-completing-read
-		       "Action to take: "
-		       gnorb-org-trigger-actions nil t)
-		      gnorb-org-trigger-actions)))
-	(root-marker (make-marker)))
-    ;; Place the marker for the relevant TODO heading.
-    (cond (agenda-p
-	   (setq root-marker
+  (let* ((agenda-p (eq major-mode 'org-agenda-mode))
+	 (root-marker
+	  (cond (agenda-p
 		 (copy-marker
-		  (org-get-at-bol 'org-hd-marker))))
-	  ((derived-mode-p 'org-mode)
-	   (move-marker root-marker (point-at-bol)))
-	  (id
-	   (save-excursion
-	     (org-id-goto id)
-	     (move-marker root-marker (point-at-bol)))))
+		  (org-get-at-bol 'org-hd-marker)))
+		((derived-mode-p 'org-mode)
+		 (save-excursion
+		   (org-back-to-heading)
+		   (point-marker)))
+		(id
+		 (save-excursion
+		   (org-id-goto id)
+		   (org-back-to-heading)
+		   (point-marker)))))
+	 (id (or id
+		 (org-with-point-at root-marker
+		   (org-id-get-create))))
+	 (action (gnorb-select-from-list
+		  (format
+		   "Trigger action on %s: "
+		   (gnorb-pretty-outline id))
+		  gnorb-org-trigger-actions)))
     (unless agenda-p
       (org-reveal))
-    ;; Query about attaching email attachments. No matter what
-    ;; happens, clear `gnorb-gnus-capture-attachments'.
-    (unwind-protect
-	(org-with-point-at root-marker
-	  (map-y-or-n-p
-	   (lambda (a)
-	     (format "Attach %s to heading? "
-		     (file-name-nondirectory a)))
-	   (lambda (a) (org-attach-attach a nil 'mv))
-	   gnorb-gnus-capture-attachments
-	   '("file" "files" "attach")))
-      (setq gnorb-gnus-capture-attachments nil))
     (cl-labels
 	((make-entry
 	  (id)
@@ -255,57 +346,113 @@ agenda. Then let the user choose an action from the value of
 	   id
 	   (plist-get gnorb-gnus-message-info :group))))
       ;; Handle our action.
-      (cond ((eq action 'note)
+      (if (fboundp action)
+	  (org-with-point-at root-marker
+	    (make-entry (org-id-get-create))
+	    (funcall action gnorb-gnus-message-info))
+	(cl-case action
+	  (note
+	   (org-with-point-at root-marker
+	     (make-entry (org-id-get-create))
+	     (call-interactively 'org-add-note)))
+	  (todo
+	   (if agenda-p
+	       (progn
+		 (org-with-point-at root-marker
+		   (make-entry (org-id-get-create)))
+		 (call-interactively 'org-agenda-todo))
 	     (org-with-point-at root-marker
 	       (make-entry (org-id-get-create))
-	       (call-interactively 'org-add-note)))
-	    ((eq action 'todo)
-	     (if agenda-p
-		 (progn
-		   (org-with-point-at root-marker
-		    (make-entry (org-id-get-create)))
-		   (call-interactively 'org-agenda-todo))
-	       (org-with-point-at root-marker
-		 (make-entry (org-id-get-create))
-		 (call-interactively 'org-todo))))
-	    ((eq action 'no-associate)
-	     nil)
-	    ((eq action 'associate)
-	     (org-with-point-at root-marker
-	       (make-entry (org-id-get-create))))
-	    ((fboundp action)
-	     (org-with-point-at root-marker
-	       (make-entry (org-id-get-create))
-	       (funcall action gnorb-gnus-message-info)))))))
+	       (call-interactively 'org-todo))))
+	  (no-associate
+	   nil)
+	  (associate
+	   (org-with-point-at root-marker
+	     (make-entry (org-id-get-create))))
+	  ;; We're going to capture a new heading
+	  ((cap-child cap-sib)
+	   (org-with-point-at root-marker
+		(setq gnorb-trigger-capture-location (point-marker)))
+	   (let ((entry
+		  ;; Pick a template.
+		  (copy-sequence (org-capture-select-template))))
+	     ;; Do surgery on that template so that it finds its
+	     ;; location using our function.
+	     (setf (nth 3 entry)
+		   `(function
+		     ,(if (eq action 'cap-child)
+			  #'gnorb-trigger-capture-child
+			#'gnorb-trigger-capture-sibling)))
+	     ;; This will likely fail horribly for capture templates
+	     ;; that aren't entries or list items.
+	     (let ((org-capture-entry entry))
+	       ;; When org-capture-entry is let-bound, the capture
+	       ;; process will use that template instead of
+	       ;; prompting the user. Also, `gnorb-registry-capture'
+	       ;; will take care of making the registry entry for us.
+	       (call-interactively 'org-capture)))))))
+    ;; Lastly, query about attaching email attachments. No matter what
+    ;; happens, clear `gnorb-gnus-capture-attachments'.
+    (unwind-protect
+	(org-with-point-at
+	    (if (memq action '(cap-child cap-sib))
+		(point)
+	      root-marker)
+	  (map-y-or-n-p
+	   (lambda (a)
+	     (format "Attach %s to heading? "
+		     (file-name-nondirectory a)))
+	   (lambda (a)
+	     (with-demoted-errors
+		 (org-attach-attach a nil 'mv)))
+	   gnorb-gnus-capture-attachments
+	   '("file" "files" "attach")))
+      (setq gnorb-gnus-capture-attachments nil))))
+
+(defun gnorb-trigger-capture-child ()
+  ;; The capture process creates a child by default
+  (org-goto-marker-or-bmk gnorb-trigger-capture-location)
+  (org-back-to-heading))
+
+(defun gnorb-trigger-capture-sibling ()
+  ;; This only works if we're not trying to create a sibling for a
+  ;; top-level heading, there appears to be no way to do that.  But in
+  ;; that case this trigger action isn't really necessary, just
+  ;; handle it with a regular capture.
+  (org-goto-marker-or-bmk gnorb-trigger-capture-location)
+  (org-up-heading-safe))
 
 (defun gnorb-pretty-outline (id &optional kw)
   "Return pretty outline path of the Org heading indicated by ID.
 
 If the KW argument is true, add the TODO keyword into the path."
-  (org-with-point-at (org-id-find id t)
-    (let ((el (org-element-at-point)))
-      (concat
-       (if kw
-	   (format "(%s): "
-		   (org-element-property :todo-keyword el))
-	 "")
-       (org-format-outline-path
-	(append
-	 (list
-	  (file-name-nondirectory
-	   (buffer-file-name
-	    (org-base-buffer (current-buffer)))))
-	 (org-get-outline-path)
-	 (list
-	  (replace-regexp-in-string
-	   org-bracket-link-regexp
-	   "\\3" (org-element-property :raw-value el)))))))))
+  (let ((pt (org-id-find id t)))
+    (if pt
+	(org-with-point-at pt
+	  (let ((el (org-element-at-point)))
+	    (concat
+	     (if kw
+		 (format "(%s): "
+			 (org-element-property :todo-keyword el))
+	       "")
+	     (org-format-outline-path
+	      (append
+	       (list
+		(file-name-nondirectory
+		 (buffer-file-name
+		  (org-base-buffer (current-buffer)))))
+	       (org-get-outline-path)
+	       (list
+		(replace-regexp-in-string
+		 org-bracket-link-regexp
+		 "\\3" (org-element-property :raw-value el))))))))
+      "[none]")))
 
 (defun gnorb-scan-links (bound &rest types)
   "Scan from point to BOUND looking for links of type in TYPES.
 
-TYPES is a list of symbols, possible values include 'bbdb, 'mail,
-and 'gnus."
+TYPES is a list of symbols, possible values include `bbdb', `mail',
+and `gnus'."
   ;; this function could be refactored somewhat -- lots of code
   ;; repetition. It also should be a little faster for when we're
   ;; scanning for gnus links only, that's a little slow. We should
@@ -346,26 +493,29 @@ message."
 So far we're checking the registry, then the groups in
 `gnorb-gnus-sent-groups'. Use search engines? Other clever
 methods?"
-  (let (candidates server-group)
+  (let (candidates server-group check)
     (setq msg-id (gnorb-bracket-message-id msg-id))
     (catch 'found
       (when gnorb-tracking-enabled
-	;; Make a big list of all the groups where this message might
-	;; conceivably be.
-	(setq candidates
-	      (append (gnus-registry-get-id-key msg-id 'group)
-		      gnorb-gnus-sent-groups))
-	(while (setq server-group (pop candidates))
-	  (when (and (stringp server-group)
-		     (not
-		      (string-match-p
-		       "\\(nnir\\|nnvirtual\\|UNKNOWN\\)"
-		       server-group))
-		     (ignore-errors
-		       (gnus-request-head msg-id server-group)))
-		(throw 'found server-group))))
-      (when (featurep 'notmuch)
-	nil))))
+	(setq candidates (gnus-registry-get-id-key msg-id 'group))
+	(if (= 1 (length candidates))
+	    (throw 'found (car candidates))
+	  (setq candidates (append candidates gnorb-gnus-sent-groups))
+	  (while (setq server-group (pop candidates))
+	    (when (and (stringp server-group)
+		       (string-match-p "+" server-group)
+		       (not
+			(string-match-p
+			 "\\(nnir\\|nnvirtual\\|UNKNOWN\\)"
+			 server-group)))
+	      (setq check
+		    (ignore-errors
+		      (gnus-request-head msg-id server-group)))
+	      (when check
+		(gnus-registry-set-id-key msg-id 'group (list server-group))
+		(gnus-registry-set-id-key msg-id 'artno (list (cdr check)))
+		(throw 'found (car check)))))))
+      nil)))
 
 (defun gnorb-collect-ids (&optional id)
   "Collect all Org IDs for a subtree.
@@ -386,6 +536,16 @@ child headings."
 ;; Common functions for extracting references and relevant headings
 ;; from the message under point. For use in gnorb-gnus.el functions.
 
+(defun gnorb-get-real-group-name (group art-no)
+  "Find the original group name of a message in a virtual or nnir
+group."
+  (cl-case (car (gnus-find-method-for-group group))
+    (nnvirtual
+     (setq group (car (nnvirtual-map-article art-no))))
+    (nnir
+     (setq group (nnir-article-group art-no))))
+  group)
+
 (defun gnorb-find-tracked-headings (headers &optional include-zombies)
   "Check HEADERS for message references and return relevant heading IDs.
 
@@ -435,10 +595,16 @@ registry be in use, and should be called after the call to
   (require 'gnorb-registry)
   (with-eval-after-load 'gnus-registry
     (add-to-list 'gnus-registry-extra-entries-precious 'gnorb-ids)
+    (add-to-list 'gnus-registry-extra-entries-precious 'artno)
     (add-to-list 'gnus-registry-track-extra 'gnorb-ids))
   (add-hook
    'gnus-started-hook
    (lambda ()
+     ;; The require may be necessary in order to get
+     ;; `gnus-user-format-function-g' defined before it's used.  That
+     ;; function is likely the first hit on gnorb-gnus, and there's no
+     ;; way to autoload it, as it is dynamically defined.
+     (require 'gnorb-gnus)
      (unless (gnus-registry-install-p)
        (user-error "Gnorb tracking requires that the Gnus registry be installed."))
      (add-hook 'org-capture-mode-hook 'gnorb-registry-capture)
